@@ -1,6 +1,3 @@
-#if __GLASGOW_HASKELL__ >= 704
-{-# LANGUAGE Safe #-}
-#endif
 {-# LANGUAGE OverloadedStrings #-}
 
 -- This template expects CPP definitions for:
@@ -63,13 +60,47 @@ module System.FilePath.MODULE_NAME.ByteString
     (
     -- * Types
     RawFilePath,
+    -- * Filename encoding
+    -- 
+    -- | When using `FilePath`, you do not usually need to care about how
+    -- it is encoded, because it is a @[Char]@ and encoding and decoding is
+    -- handled by IO actions as needed. Unfortunately the situation is more
+    -- complicated when using `RawFilePath`.
+    --
+    -- It's natural to enable `OverloadedStrings` and use it to construct
+    -- a `RawFilePath`, eg @"foo" '</>' "bar"@. A gotcha though is that
+    -- any non-ascii characters will be truncated to 8 bits. That is not a
+    -- limitation of this library, but of the `IsString` implementation 
+    -- of `ByteString`.
+    --
+    -- Posix filenames do not have any defined encoding. This library
+    -- assumes that whatever encoding may be used for a `RawFilePath`,
+    -- it is compatable with ASCII. In particular, 0x2F (/) is always
+    -- a path separator, and 0x2E (.) is assumed to be an extension
+    -- separator. All encodings in common use are compatible with ASCII,
+    -- and unix tools have always made similar assumptions,
+    -- so this is unlikely to be a problem, unless you are dealing with
+    -- EBCDIC or similar historical oddities.
+    --
+    -- Windows's API expects filenames to be encoded with UTF-16.
+    -- This is especially problimatic when using OverloadedStrings
+    -- since a ByteString "bar" is not a valid encoding for a
+    -- Windows filename (but "b\\0a\\0r\\0" is). To avoid this problem,
+    -- and to simplify the implementation,
+    -- `RawFilePath` is assumed to be encoded with UTF-8 (not UTF-16)
+    -- when this library is used on Windows.
+    -- There are not currently any libraries for Windows that use
+    -- `RawFilePath`, so you will probably need to convert them back to 
+    -- `FilePath` in order to do IO in any case.
+    encodeFilePath,
+    decodeFilePath,
     -- * Separator predicates
     pathSeparator, pathSeparators, isPathSeparator,
     searchPathSeparator, isSearchPathSeparator,
     extSeparator, isExtSeparator,
 
     -- * @$PATH@ methods
-    --splitSearchPath, getSearchPath,
+    splitSearchPath, getSearchPath,
 
     -- * Extension functions
     splitExtension,
@@ -95,19 +126,29 @@ module System.FilePath.MODULE_NAME.ByteString
     dropTrailingPathSeparator,
 
     -- * File name manipulations
-    --normalise, equalFilePath,
-    --makeRelative,
+    normalise, equalFilePath,
+    makeRelative,
     isRelative, isAbsolute,
-    isValid, -- makeValid
+    isValid, makeValid
     )
     where
 
 import Data.ByteString (ByteString)
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
-import Data.Char(ord, chr, toUpper, isAsciiLower, isAsciiUpper)
+import Data.Char(ord, chr, toUpper, toLower, isAsciiLower, isAsciiUpper)
 import Data.Maybe(isJust)
 import Data.Word(Word8)
+#ifdef mingw32_HOST_OS
+import qualified Data.ByteString.UTF8 as UTF8
+#else
+import qualified GHC.Foreign as GHC
+import qualified GHC.IO.Encoding as Encoding
+import System.IO.Unsafe (unsafePerformIO)
+#endif
+import System.Environment (getEnv)
+import Data.Semigroup
+import Prelude
 
 #ifndef mingw32_HOST_OS
 -- Import from unix, rather than redefining, so users who import both
@@ -123,6 +164,45 @@ infixr 7  <.>, -<.>
 infixr 5  </>
 
 
+-- | Convert from FilePath to RawFilePath.
+--
+-- When run on Unix, this applies the filesystem encoding
+-- (see `Encoding.getFileSystemEncoding`).
+--
+-- When run on Windows, this encodes as UTF-8.
+encodeFilePath :: FilePath -> RawFilePath
+#ifdef mingw32_HOST_OS
+encodeFilePath = UTF8.fromString
+#else
+encodeFilePath = B.pack . map (fromIntegral . fromEnum) . encodeFilePath'
+
+{-# NOINLINE encodeFilePath' #-}
+encodeFilePath' :: FilePath -> String
+encodeFilePath' f = unsafePerformIO $ do
+	enc <- Encoding.getFileSystemEncoding
+	GHC.withCString enc f (GHC.peekCString Encoding.char8)
+#endif
+
+-- | Convert from RawFilePath to FilePath
+--
+-- When run on Unix, this applies the filesystem encoding
+-- (see `Encoding.getFileSystemEncoding`).
+--
+-- When run on Windows, this decodes UTF-8.
+decodeFilePath :: RawFilePath -> FilePath
+#ifdef mingw32_HOST_OS
+decodeFilePath = UTF8.toString
+#else
+decodeFilePath = decodeFilePath' . map (toEnum . fromIntegral) . B.unpack
+
+{-# NOINLINE decodeFilePath' #-}
+decodeFilePath' :: String -> FilePath
+decodeFilePath' s = unsafePerformIO $ do
+	enc <- Encoding.getFileSystemEncoding
+	GHC.withCString Encoding.char8 s (GHC.peekCString enc)
+#endif
+
+
 ---------------------------------------------------------------------
 -- Platform Abstraction Methods (private)
 
@@ -191,7 +271,6 @@ extSeparator = fromIntegral (ord '.')
 isExtSeparator :: Word8 -> Bool
 isExtSeparator = (== extSeparator)
 
-{-
 
 ---------------------------------------------------------------------
 -- Path methods (environment $PATH)
@@ -203,28 +282,36 @@ isExtSeparator = (== extSeparator)
 --   Follows the recommendations in
 --   <http://www.opengroup.org/onlinepubs/009695399/basedefs/xbd_chap08.html>
 --
--- > untested Posix:   splitSearchPath "File1:File2:File3"  == ["File1","File2","File3"]
--- > untested Posix:   splitSearchPath "File1::File2:File3" == ["File1",".","File2","File3"]
--- > untested Windows: splitSearchPath "File1;File2;File3"  == ["File1","File2","File3"]
--- > untested Windows: splitSearchPath "File1;;File2;File3" == ["File1","File2","File3"]
--- > untested Windows: splitSearchPath "File1;\"File2\";File3" == ["File1","File2","File3"]
-splitSearchPath :: String -> [FilePath]
+-- > Posix:   splitSearchPath "File1:File2:File3"  == ["File1","File2","File3"]
+-- > Posix:   splitSearchPath "File1::File2:File3" == ["File1",".","File2","File3"]
+-- > Windows: splitSearchPath "File1;File2;File3"  == ["File1","File2","File3"]
+-- > Windows: splitSearchPath "File1;;File2;File3" == ["File1","File2","File3"]
+-- > Windows: splitSearchPath "File1;\"File2\";File3" == ["File1","File2","File3"]
+splitSearchPath :: ByteString -> [RawFilePath]
 splitSearchPath = f
-    where
-    f xs = case break isSearchPathSeparator xs of
-           (pre, []    ) -> g pre
-           (pre, _:post) -> g pre ++ f post
+  where
+	f x = case B.break isSearchPathSeparator x of
+		(pre, post)
+			| B.null post -> g pre
+			| otherwise -> g pre ++ f (B.drop 1 post)
 
-    g "" = ["." | isPosix]
-    g ('\"':x@(_:_)) | isWindows && last x == '\"' = [init x]
-    g x = [x]
+	g "" = ["." | isPosix]
+	g x
+		| isWindows = case B.uncons x of
+			Just (q, x') | q == quote ->
+				case B.unsnoc x' of
+					Just (x'', q') | q' == quote -> [x'']
+					_ -> [x]
+			_  -> [x]
+		| otherwise = [x]
+	
+	quote = fromIntegral (ord '"')
 
 
--- | Get a list of 'FilePath's in the $PATH variable.
-getSearchPath :: IO [FilePath]
-getSearchPath = fmap splitSearchPath (getEnv "PATH")
+-- | Get a list of 'RawFilePath's in the $PATH variable.
+getSearchPath :: IO [RawFilePath]
+getSearchPath = fmap (splitSearchPath . encodeFilePath) (getEnv "PATH")
 
--}
 
 ---------------------------------------------------------------------
 -- Extension methods
@@ -243,11 +330,11 @@ getSearchPath = fmap splitSearchPath (getEnv "PATH")
 -- > splitExtension "file/path.txt/" == ("file/path.txt/","")
 splitExtension :: RawFilePath -> (ByteString, ByteString)
 splitExtension x = if B.null nameDot
-        then (x,mempty)
-        else (dir <> B.init nameDot, extSeparator `B.cons` ext)
-    where
-        (dir,file) = splitFileName_ x
-        (nameDot,ext) = B.breakEnd isExtSeparator file
+	then (x,mempty)
+	else (dir <> B.init nameDot, extSeparator `B.cons` ext)
+  where
+	(dir,file) = splitFileName_ x
+	(nameDot,ext) = B.breakEnd isExtSeparator file
 
 -- | Get the extension of a file, returns @\"\"@ for no extension, @.ext@ otherwise.
 --
@@ -305,12 +392,13 @@ dropExtension = fst . splitExtension
 -- > Windows: addExtension "\\\\share" ".txt" == "\\\\share\\.txt"
 addExtension :: RawFilePath -> ByteString -> RawFilePath
 addExtension file ext = case B.uncons ext of
-        Nothing -> file
-        Just (x,_xs) -> joinDrive a $
-                if isExtSeparator x then b <> ext
-                else b <> (extSeparator `B.cons` ext)
-    where
-        (a,b) = splitDrive file
+	Nothing -> file
+	Just (x,_xs) -> joinDrive a $
+		if isExtSeparator x
+			then b <> ext
+			else b <> (extSeparator `B.cons` ext)
+  where
+	(a,b) = splitDrive file
 
 -- | Does the given filename have an extension?
 --
@@ -332,9 +420,9 @@ hasExtension = B.any isExtSeparator . takeFileName
 isExtensionOf :: ByteString -> RawFilePath -> Bool
 isExtensionOf ext = B.isSuffixOf ext' . takeExtensions
   where
-    ext' = case B.uncons ext of
-        Just (h, _) | isExtSeparator h -> ext
-        _ -> extSeparator `B.cons` ext
+	ext' = case B.uncons ext of
+		Just (h, _) | isExtSeparator h -> ext
+		_ -> extSeparator `B.cons` ext
 
 -- | Drop the given extension from a FilePath, and the @\".\"@ preceding it.
 --   Returns 'Nothing' if the FilePath does not have the given extension, or
@@ -354,12 +442,12 @@ isExtensionOf ext = B.isSuffixOf ext' . takeExtensions
 -- > stripExtension ""     x          == Just x
 stripExtension :: ByteString -> RawFilePath -> Maybe RawFilePath
 stripExtension ext path = case B.uncons ext of
-        Nothing -> Just path
-        Just (x, _) -> 
-                let dotExt = if isExtSeparator x
-                        then ext
-                        else extSeparator `B.cons` ext
-                in B.stripSuffix dotExt path
+	Nothing -> Just path
+	Just (x, _) -> 
+		let dotExt = if isExtSeparator x
+			then ext
+			else extSeparator `B.cons` ext
+		in B.stripSuffix dotExt path
 
 
 -- | Split on all extensions.
@@ -371,9 +459,9 @@ stripExtension ext path = case B.uncons ext of
 -- > splitExtensions "file.tar.gz" == ("file",".tar.gz")
 splitExtensions :: RawFilePath -> (RawFilePath, ByteString)
 splitExtensions x = (a <> c, d)
-    where
-        (a,b) = splitFileName_ x
-        (c,d) = B.break isExtSeparator b
+  where
+	(a,b) = splitFileName_ x
+	(c,d) = B.break isExtSeparator b
 
 -- | Drop all extensions.
 --
@@ -439,64 +527,67 @@ splitDrive x = (B.empty,x)
 
 addSlash :: RawFilePath -> RawFilePath -> (RawFilePath, RawFilePath)
 addSlash a xs = (a <> c,d)
-    where (c,d) = B.span isPathSeparator xs
+  where
+	(c,d) = B.span isPathSeparator xs
 
 -- See [1].
 -- "\\?\D:\<path>" or "\\?\UNC\<server>\<share>"
 readDriveUNC :: RawFilePath -> Maybe (RawFilePath, RawFilePath)
 readDriveUNC p = do
-        (s1, p') <- B.uncons p
-        (s2, p'') <- B.uncons p'
-        (q, p''') <- B.uncons p''
-        (s3, rest) <- B.uncons p'''
-        if q == fromIntegral (ord '?') && all isPathSeparator [s1,s2,s3]
-                then case breakunc rest of
-                        Just (unc,r) -> 
-                                let (a,b) = readDriveShareName r
-                                in Just (B.pack [s1,s2,q,s3] <> unc <> a, b)
-                        Nothing -> case readDriveLetter rest of
-                                -- Extended-length path.
-                                Just (a,b) -> Just (B.pack [s1,s2,q,s3] <> a, b)
-                                Nothing -> Nothing
-                else Nothing
+	(s1, p') <- B.uncons p
+	(s2, p'') <- B.uncons p'
+	(q, p''') <- B.uncons p''
+	(s3, rest) <- B.uncons p'''
+	if q == fromIntegral (ord '?') && all isPathSeparator [s1,s2,s3]
+		then case breakunc rest of
+			Just (unc,r) -> 
+				let (a,b) = readDriveShareName r
+				in Just (B.pack [s1,s2,q,s3] <> unc <> a, b)
+			Nothing -> case readDriveLetter rest of
+				-- Extended-length path.
+				Just (a,b) -> Just (B.pack [s1,s2,q,s3] <> a, b)
+				Nothing -> Nothing
+		else Nothing
   where
-        breakunc b = do
-                (u, b') <- B.uncons b
-                (n, b'') <- B.uncons b'
-                (c, b''') <- B.uncons b''
-                (s, rest) <- B.uncons b'''
-                let isup v ch = toUpper (chr (fromIntegral v)) == ch
-                if isPathSeparator s && isup u 'U' && isup n 'N' && isup c 'C'
-                        then Just (B.take 4 b, rest)
-                        else Nothing
+	breakunc b = do
+		(u, b') <- B.uncons b
+		(n, b'') <- B.uncons b'
+		(c, b''') <- B.uncons b''
+		(s, rest) <- B.uncons b'''
+		let isup v ch = toUpper (chr (fromIntegral v)) == ch
+		if isPathSeparator s && isup u 'U' && isup n 'N' && isup c 'C'
+			then Just (B.take 4 b, rest)
+			else Nothing
 
 {- c:\ -}
 readDriveLetter :: RawFilePath -> Maybe (RawFilePath, RawFilePath)
 readDriveLetter p = case B.uncons p of
-        Just (x, t) | isLetter x -> case B.uncons t of
-                Just (c, t') | c == colon -> case B.uncons t' of
-                        Just (y, _) | isPathSeparator y ->
-                                Just $ addSlash (B.pack [x, colon]) t'
-                        _ -> Just (B.pack [x, colon], t')
-                _ -> Nothing
-        _ -> Nothing
-  where colon = fromIntegral (ord ':')
+	Just (x, t) | isLetter x -> case B.uncons t of
+		Just (c, t') | c == colon -> case B.uncons t' of
+			Just (y, _) | isPathSeparator y ->
+				Just $ addSlash (B.pack [x, colon]) t'
+			_ -> Just (B.pack [x, colon], t')
+		_ -> Nothing
+	_ -> Nothing
+  where
+	colon = fromIntegral (ord ':')
 
 {- \\sharename\ -}
 readDriveShare :: ByteString -> Maybe (RawFilePath, RawFilePath)
 readDriveShare p = do
-        (s1, p') <- B.uncons p
-        (s2, p'') <- B.uncons p'
-        let (a,b) = readDriveShareName p''
-        if isPathSeparator s1 && isPathSeparator s2
-                then Just (s1 `B.cons` s2 `B.cons` a,b)
-                else Nothing
+	(s1, p') <- B.uncons p
+	(s2, p'') <- B.uncons p'
+	let (a,b) = readDriveShareName p''
+	if isPathSeparator s1 && isPathSeparator s2
+		then Just (s1 `B.cons` s2 `B.cons` a,b)
+		else Nothing
 
 {- assume you have already seen \\ -}
 {- share\bob -> "share\", "bob" -}
 readDriveShareName :: ByteString -> (RawFilePath, RawFilePath)
 readDriveShareName name = addSlash a b
-    where (a,b) = B.break isPathSeparator name
+  where
+	(a,b) = B.break isPathSeparator name
 
 
 -- | Join a drive and the rest of the path.
@@ -558,8 +649,8 @@ isDrive x = not (B.null x) && B.null (dropDrive x)
 -- > Windows: splitFileName "c:" == ("c:","")
 splitFileName :: RawFilePath -> (ByteString, ByteString)
 splitFileName x = (if B.null dir then "./" else dir, name)
-    where
-        (dir, name) = splitFileName_ x
+  where
+	(dir, name) = splitFileName_ x
 
 -- version of splitFileName where, if the FilePath has no directory
 -- component, the returned directory is "" rather than "./".  This
@@ -569,9 +660,9 @@ splitFileName x = (if B.null dir then "./" else dir, name)
 -- e.g. replaceFileName.
 splitFileName_ :: RawFilePath -> (ByteString, ByteString)
 splitFileName_ x = (drv <> dir, file)
-    where
-        (drv,pth) = splitDrive x
-        (dir,file) = B.breakEnd isPathSeparator pth
+  where
+	(drv,pth) = splitDrive x
+	(dir,file) = B.breakEnd isPathSeparator pth
 
 -- | Set the filename.
 --
@@ -621,9 +712,9 @@ takeBaseName = dropExtension . takeFileName
 -- > Valid x => replaceBaseName x (takeBaseName x) == x
 replaceBaseName :: RawFilePath -> ByteString -> RawFilePath
 replaceBaseName pth nam = combineAlways a (nam <.> ext)
-    where
-        (a,b) = splitFileName_ pth
-        ext = takeExtension b
+  where
+	(a,b) = splitFileName_ pth
+	ext = takeExtension b
 
 -- | Is an item either a directory or the last character a path separator?
 --
@@ -631,14 +722,14 @@ replaceBaseName pth nam = combineAlways a (nam <.> ext)
 -- > hasTrailingPathSeparator "test/" == True
 hasTrailingPathSeparator :: RawFilePath -> Bool
 hasTrailingPathSeparator x = case B.unsnoc x of
-        Nothing -> False
-        Just (_i, l) -> isPathSeparator l
+	Nothing -> False
+	Just (_i, l) -> isPathSeparator l
 
 
 hasLeadingPathSeparator :: RawFilePath -> Bool
 hasLeadingPathSeparator x = case B.uncons x of
-        Nothing -> False
-        Just (h,_t) -> isPathSeparator h
+	Nothing -> False
+	Just (h,_t) -> isPathSeparator h
 
 -- | Add a trailing file path separator if one is not already present.
 --
@@ -647,8 +738,8 @@ hasLeadingPathSeparator x = case B.uncons x of
 -- > Posix:    addTrailingPathSeparator "test/rest" == "test/rest/"
 addTrailingPathSeparator :: RawFilePath -> RawFilePath
 addTrailingPathSeparator x
-        | hasTrailingPathSeparator x = x
-        | otherwise = B.snoc x pathSeparator
+	| hasTrailingPathSeparator x = x
+	| otherwise = B.snoc x pathSeparator
 
 
 -- | Remove any trailing path separators
@@ -658,12 +749,11 @@ addTrailingPathSeparator x
 -- > Windows:  dropTrailingPathSeparator "\\" == "\\"
 -- > Posix:    not (hasTrailingPathSeparator (dropTrailingPathSeparator x)) || isDrive x
 dropTrailingPathSeparator :: RawFilePath -> RawFilePath
-dropTrailingPathSeparator x =
-    if hasTrailingPathSeparator x && not (isDrive x)
-    then let x' = dropWhileEnd isPathSeparator x
-         in if B.null x' then B.singleton (B.last x) else x'
-    else x
-
+dropTrailingPathSeparator x
+	| hasTrailingPathSeparator x && not (isDrive x) = 
+		let x' = dropWhileEnd isPathSeparator x
+		in if B.null x' then B.singleton (B.last x) else x'
+	| otherwise = x
 
 -- | Get the directory name, move up one level.
 --
@@ -684,24 +774,27 @@ takeDirectory = dropTrailingPathSeparator . dropFileName
 -- | Set the directory, keeping the filename the same.
 --
 -- > replaceDirectory "root/file.ext" "/directory/" == "/directory/file.ext"
--- > untested Valid x => replaceDirectory x (takeDirectory x) `equalFilePath` x
+-- > Valid x => replaceDirectory x (takeDirectory x) `equalFilePath` x
 replaceDirectory :: RawFilePath -> ByteString -> RawFilePath
 replaceDirectory x dir = combineAlways dir (takeFileName x)
 
+{-# INLINE combine #-}
 -- | An alias for '</>'.
 combine :: RawFilePath -> RawFilePath -> RawFilePath
-combine a b | hasLeadingPathSeparator b || hasDrive b = b
-            | otherwise = combineAlways a b
+combine a b
+	| hasLeadingPathSeparator b || (not isPosix && hasDrive b) = b
+	| otherwise = combineAlways a b
 
 -- | Combine two paths, assuming rhs is NOT absolute.
 combineAlways :: RawFilePath -> RawFilePath -> RawFilePath
-combineAlways a b | B.null a = b
-                  | B.null b = a
-                  | hasTrailingPathSeparator a = a <> b
-                  | isWindows && B.length a == 2
-                      && B.last a == fromIntegral (ord ':')
-                      && isLetter (B.head a) = a <> b
-                  | otherwise = a <> B.singleton pathSeparator <> b
+combineAlways a b
+	| B.null a = b
+	| B.null b = a
+	| hasTrailingPathSeparator a = a <> b
+	| isWindows && B.length a == 2
+		&& B.last a == fromIntegral (ord ':')
+		&& isLetter (B.head a) = a <> b
+	| otherwise = a <> B.singleton pathSeparator <> b
 
 -- | Combine two paths with a path separator.
 --   If the second path starts with a path separator or a drive letter, then it returns the second.
@@ -711,7 +804,7 @@ combineAlways a b | B.null a = b
 -- > Posix:   "/directory" </> "file.ext" == "/directory/file.ext"
 -- > Windows: "/directory" </> "file.ext" == "/directory\\file.ext"
 -- >          "directory" </> "/file.ext" == "/file.ext"
--- > untested Valid x => (takeDirectory x </> takeFileName x) `equalFilePath` x
+-- > Valid x => (takeDirectory x </> takeFileName x) `equalFilePath` x
 --
 --   Combined:
 --
@@ -758,15 +851,15 @@ combineAlways a b | B.null a = b
 -- > Posix:   splitPath "/file/test" == ["/","file/","test"]
 splitPath :: RawFilePath -> [RawFilePath]
 splitPath x = [drive | not (B.null drive)] ++ f path
-    where
-        (drive,path) = splitDrive x
+  where
+	(drive,path) = splitDrive x
 
-        f y
-                | B.null y = []
-                | otherwise = (a<>c) : f d
-            where
-                (a,b) = B.break isPathSeparator y
-                (c,d) = B.span  isPathSeparator b
+	f y
+		| B.null y = []
+		| otherwise = (a<>c) : f d
+	  where
+		(a,b) = B.break isPathSeparator y
+		(c,d) = B.span  isPathSeparator b
 
 -- | Just as 'splitPath', but don't add the trailing slashes to each element.
 --
@@ -774,7 +867,7 @@ splitPath x = [drive | not (B.null drive)] ++ f path
 -- >          splitDirectories "test/file" == ["test","file"]
 -- >          splitDirectories "/test/file" == ["/","test","file"]
 -- > Windows: splitDirectories "C:\\test\\file" == ["C:\\", "test", "file"]
--- > untested          Valid x => joinPath (splitDirectories x) `equalFilePath` x
+-- >          Valid x => joinPath (splitDirectories x) `equalFilePath` x
 -- >          splitDirectories "" == []
 -- > Windows: splitDirectories "C:\\test\\\\\\file" == ["C:\\", "test", "file"]
 -- >          splitDirectories "/test///file" == ["/","test","file"]
@@ -792,9 +885,6 @@ joinPath :: [RawFilePath] -> RawFilePath
 -- Note that this definition on c:\\c:\\, join then split will give c:\\.
 joinPath = foldr combine mempty
 
-{-
-
-
 ---------------------------------------------------------------------
 -- File name manipulators
 
@@ -803,19 +893,20 @@ joinPath = foldr combine mempty
 --   first this has a much better chance of working.
 --   Note that this doesn't follow symlinks or DOSNAM~1s.
 --
--- > untested          x == y ==> equalFilePath x y
--- > untested          normalise x == normalise y ==> equalFilePath x y
--- > untested          equalFilePath "foo" "foo/"
--- > untested          not (equalFilePath "foo" "/foo")
--- > untested Posix:   not (equalFilePath "foo" "FOO")
--- > untested Windows: equalFilePath "foo" "FOO"
--- > untested Windows: not (equalFilePath "C:" "C:/")
-equalFilePath :: FilePath -> FilePath -> Bool
+-- >          x == y ==> equalFilePath x y
+-- >          normalise x == normalise y ==> equalFilePath x y
+-- >          equalFilePath "foo" "foo/"
+-- >          not (equalFilePath "foo" "/foo")
+-- > Posix:   not (equalFilePath "foo" "FOO")
+-- > Windows: equalFilePath "foo" "FOO"
+-- > Windows: not (equalFilePath "C:" "C:/")
+equalFilePath :: RawFilePath -> RawFilePath -> Bool
 equalFilePath a b = f a == f b
-    where
-        f x | isWindows = dropTrailingPathSeparator $ map toLower $ normalise x
-            | otherwise = dropTrailingPathSeparator $ normalise x
-
+  where
+	f x
+		| isWindows = dropTrailingPathSeparator $ encodeFilePath $
+			map toLower $ decodeFilePath $ normalise x
+		| otherwise = dropTrailingPathSeparator $ normalise x
 
 -- | Contract a filename, based on a relative path. Note that the resulting path
 --   will never introduce @..@ paths, as the presence of symlinks means @..\/b@
@@ -825,42 +916,46 @@ equalFilePath a b = f a == f b
 --   The corresponding @makeAbsolute@ function can be found in
 --   @System.Directory@.
 --
--- > untested          makeRelative "/directory" "/directory/file.ext" == "file.ext"
--- > untested          Valid x => makeRelative (takeDirectory x) x `equalFilePath` takeFileName x
--- > untested          makeRelative x x == "."
--- > untested          Valid x y => equalFilePath x y || (isRelative x && makeRelative y x == x) || equalFilePath (y </> makeRelative y x) x
--- > untested Windows: makeRelative "C:\\Home" "c:\\home\\bob" == "bob"
--- > untested Windows: makeRelative "C:\\Home" "c:/home/bob" == "bob"
--- > untested Windows: makeRelative "C:\\Home" "D:\\Home\\Bob" == "D:\\Home\\Bob"
--- > untested Windows: makeRelative "C:\\Home" "C:Home\\Bob" == "C:Home\\Bob"
--- > untested Windows: makeRelative "/Home" "/home/bob" == "bob"
--- > untested Windows: makeRelative "/" "//" == "//"
--- > untested Posix:   makeRelative "/Home" "/home/bob" == "/home/bob"
--- > untested Posix:   makeRelative "/home/" "/home/bob/foo/bar" == "bob/foo/bar"
--- > untested Posix:   makeRelative "/fred" "bob" == "bob"
--- > untested Posix:   makeRelative "/file/test" "/file/test/fred" == "fred"
--- > untested Posix:   makeRelative "/file/test" "/file/test/fred/" == "fred/"
--- > untested Posix:   makeRelative "some/path" "some/path/a/b/c" == "a/b/c"
-makeRelative :: FilePath -> FilePath -> FilePath
+-- >          makeRelative "/directory" "/directory/file.ext" == "file.ext"
+-- >          Valid x => makeRelative (takeDirectory x) x `equalFilePath` takeFileName x
+-- >          makeRelative x x == "."
+-- >          Valid x y => equalFilePath x y || (isRelative x && makeRelative y x == x) || equalFilePath (y </> makeRelative y x) x
+-- > Windows: makeRelative "C:\\Home" "c:\\home\\bob" == "bob"
+-- > Windows: makeRelative "C:\\Home" "c:/home/bob" == "bob"
+-- > Windows: makeRelative "C:\\Home" "D:\\Home\\Bob" == "D:\\Home\\Bob"
+-- > Windows: makeRelative "C:\\Home" "C:Home\\Bob" == "C:Home\\Bob"
+-- > Windows: makeRelative "/Home" "/home/bob" == "bob"
+-- > Windows: makeRelative "/" "//" == "//"
+-- > Posix:   makeRelative "/Home" "/home/bob" == "/home/bob"
+-- > Posix:   makeRelative "/home/" "/home/bob/foo/bar" == "bob/foo/bar"
+-- > Posix:   makeRelative "/fred" "bob" == "bob"
+-- > Posix:   makeRelative "/file/test" "/file/test/fred" == "fred"
+-- > Posix:   makeRelative "/file/test" "/file/test/fred/" == "fred/"
+-- > Posix:   makeRelative "some/path" "some/path/a/b/c" == "a/b/c"
+makeRelative :: RawFilePath -> RawFilePath -> RawFilePath
 makeRelative root path
- | equalFilePath root path = "."
- | takeAbs root /= takeAbs path = path
- | otherwise = f (dropAbs root) (dropAbs path)
-    where
-        f "" y = dropWhile isPathSeparator y
-        f x y = let (x1,x2) = g x
-                    (y1,y2) = g y
-                in if equalFilePath x1 y1 then f x2 y2 else path
+	| equalFilePath root path = "."
+	| takeAbs root /= takeAbs path = path
+	| otherwise = f (dropAbs root) (dropAbs path)
+  where
+	f "" y = B.dropWhile isPathSeparator y
+	f x y =
+		let (x1,x2) = g x
+		    (y1,y2) = g y
+		in if equalFilePath x1 y1 then f x2 y2 else path
 
-        g x = (dropWhile isPathSeparator a, dropWhile isPathSeparator b)
-            where (a,b) = break isPathSeparator $ dropWhile isPathSeparator x
+	g x = (B.dropWhile isPathSeparator a, B.dropWhile isPathSeparator b)
+	  where
+		(a,b) = B.break isPathSeparator $ B.dropWhile isPathSeparator x
 
-        -- on windows, need to drop '/' which is kind of absolute, but not a drive
-        dropAbs x | hasLeadingPathSeparator x && not (hasDrive x) = tail x
-        dropAbs x = dropDrive x
+	-- on windows, need to drop '/' which is kind of absolute, but not a drive
+	dropAbs x | hasLeadingPathSeparator x && not (hasDrive x) = B.tail x
+	dropAbs x = dropDrive x
 
-        takeAbs x | hasLeadingPathSeparator x && not (hasDrive x) = [pathSeparator]
-        takeAbs x = map (\y -> if isPathSeparator y then pathSeparator else toLower y) $ takeDrive x
+	takeAbs x | hasLeadingPathSeparator x && not (hasDrive x) = B.singleton pathSeparator
+	-- A Windows drive letter is an ascii character, so it's safe to
+	-- operate on the ByteString containing it using B8.
+	takeAbs x = B8.map (\y -> if isPathSeparator (fromIntegral (ord y)) then chr (fromIntegral pathSeparator) else toLower y) $ takeDrive x
 
 -- | Normalise a file
 --
@@ -870,75 +965,85 @@ makeRelative root path
 --
 -- * .\/ -> \"\"
 --
--- > untested Posix:   normalise "/file/\\test////" == "/file/\\test/"
--- > untested Posix:   normalise "/file/./test" == "/file/test"
--- > untested Posix:   normalise "/test/file/../bob/fred/" == "/test/file/../bob/fred/"
--- > untested Posix:   normalise "../bob/fred/" == "../bob/fred/"
--- > untested Posix:   normalise "./bob/fred/" == "bob/fred/"
--- > untested Windows: normalise "c:\\file/bob\\" == "C:\\file\\bob\\"
--- > untested Windows: normalise "c:\\" == "C:\\"
--- > untested Windows: normalise "C:.\\" == "C:"
--- > untested Windows: normalise "\\\\server\\test" == "\\\\server\\test"
--- > untested Windows: normalise "//server/test" == "\\\\server\\test"
--- > untested Windows: normalise "c:/file" == "C:\\file"
--- > untested Windows: normalise "/file" == "\\file"
--- > untested Windows: normalise "\\" == "\\"
--- > untested Windows: normalise "/./" == "\\"
--- > untested          normalise "." == "."
--- > untested Posix:   normalise "./" == "./"
--- > untested Posix:   normalise "./." == "./"
--- > untested Posix:   normalise "/./" == "/"
--- > untested Posix:   normalise "/" == "/"
--- > untested Posix:   normalise "bob/fred/." == "bob/fred/"
--- > untested Posix:   normalise "//home" == "/home"
-normalise :: FilePath -> FilePath
-normalise path = result ++ [pathSeparator | addPathSeparator]
+-- > Posix:   normalise "/file/\\test////" == "/file/\\test/"
+-- > Posix:   normalise "/file/./test" == "/file/test"
+-- > Posix:   normalise "/test/file/../bob/fred/" == "/test/file/../bob/fred/"
+-- > Posix:   normalise "../bob/fred/" == "../bob/fred/"
+-- > Posix:   normalise "./bob/fred/" == "bob/fred/"
+-- > Windows: normalise "c:\\file/bob\\" == "C:\\file\\bob\\"
+-- > Windows: normalise "c:\\" == "C:\\"
+-- > Windows: normalise "C:.\\" == "C:"
+-- > Windows: normalise "\\\\server\\test" == "\\\\server\\test"
+-- > Windows: normalise "//server/test" == "\\\\server\\test"
+-- > Windows: normalise "c:/file" == "C:\\file"
+-- > Windows: normalise "/file" == "\\file"
+-- > Windows: normalise "\\" == "\\"
+-- > Windows: normalise "/./" == "\\"
+-- >          normalise "." == "."
+-- > Posix:   normalise "./" == "./"
+-- > Posix:   normalise "./." == "./"
+-- > Posix:   normalise "/./" == "/"
+-- > Posix:   normalise "/" == "/"
+-- > Posix:   normalise "bob/fred/." == "bob/fred/"
+-- > Posix:   normalise "//home" == "/home"
+normalise :: RawFilePath -> RawFilePath
+normalise path
+	| addPathSeparator = result <> B.singleton pathSeparator
+	| otherwise = result
     where
-        (drv,pth) = splitDrive path
-        result = joinDrive' (normaliseDrive drv) (f pth)
+	(drv,pth) = splitDrive path
+	result = joinDrive' (normaliseDrive drv) (f pth)
 
-        joinDrive' "" "" = "."
-        joinDrive' d p = joinDrive d p
+	joinDrive' "" "" = "."
+	joinDrive' d p = joinDrive d p
 
-        addPathSeparator = isDirPath pth
-            && not (hasTrailingPathSeparator result)
-            && not (isRelativeDrive drv)
+	addPathSeparator = isDirPath pth
+		&& not (hasTrailingPathSeparator result)
+		&& not (isRelativeDrive drv)
 
-        isDirPath xs = hasTrailingPathSeparator xs
-            || not (null xs) && last xs == '.' && hasTrailingPathSeparator (init xs)
+	isDirPath xs = hasTrailingPathSeparator xs
+		|| not (B.null xs) && B.last xs == extSeparator
+		&& hasTrailingPathSeparator (B.init xs)
 
-        f = joinPath . dropDots . propSep . splitDirectories
+	f :: RawFilePath -> RawFilePath
+	f = joinPath . dropDots . propSep . splitDirectories
 
-        propSep (x:xs) | all isPathSeparator x = [pathSeparator] : xs
-                       | otherwise = x : xs
-        propSep [] = []
+	propSep :: [RawFilePath] -> [RawFilePath]
+	propSep (x:xs)
+		| B.all isPathSeparator x = B.singleton pathSeparator : xs
+		| otherwise = x : xs
+	propSep [] = []
 
-        dropDots = filter ("." /=)
+	dropDots :: [RawFilePath] -> [RawFilePath]
+	dropDots = filter ("." /=)
 
-normaliseDrive :: FilePath -> FilePath
+normaliseDrive :: RawFilePath -> RawFilePath
 normaliseDrive "" = ""
-normaliseDrive _ | isPosix = [pathSeparator]
-normaliseDrive drive = if isJust $ readDriveLetter x2
-                       then map toUpper x2
-                       else x2
-    where
-        x2 = map repSlash drive
+normaliseDrive _ | isPosix = B.singleton pathSeparator
+normaliseDrive drive
+	| isJust (readDriveLetter x2) = upcasedriveletter x2
+	| otherwise = x2
+  where
+	x2 = B.map repSlash drive
 
-        repSlash x = if isPathSeparator x then pathSeparator else x
+	repSlash x = if isPathSeparator x then pathSeparator else x
 
--}
+	-- A Windows drive letter is an ascii character, so it's safe to
+	-- operate on the ByteString containing it using B8.
+	upcasedriveletter = B8.map toUpper
 
 -- Information for validity functions on Windows. See [1].
 isBadCharacter :: Word8 -> Bool
 isBadCharacter x = x >= 0 && x <= 31 || x `elem` l
   where
-        l = map (fromIntegral . ord) ":*?><|\""
+	l = map (fromIntegral . ord) ":*?><|\""
 
 badElements :: [FilePath]
 badElements =
-    ["CON","PRN","AUX","NUL","CLOCK$"
-    ,"COM1","COM2","COM3","COM4","COM5","COM6","COM7","COM8","COM9"
-    ,"LPT1","LPT2","LPT3","LPT4","LPT5","LPT6","LPT7","LPT8","LPT9"]
+	["CON","PRN","AUX","NUL","CLOCK$"
+	,"COM1","COM2","COM3","COM4","COM5","COM6","COM7","COM8","COM9"
+	,"LPT1","LPT2","LPT3","LPT4","LPT5","LPT6","LPT7","LPT8","LPT9"
+	]
 
 -- | Is a RawFilePath valid, i.e. could you create a file like it? This function checks for invalid names,
 --   and invalid characters, but does not check if length limits are exceeded, as these are typically
@@ -962,58 +1067,57 @@ badElements =
 -- > Windows: isValid " nul.txt" == True
 isValid :: RawFilePath -> Bool
 isValid path
-        | B.null path = False
-        | B.elem 0 path = False
-        | isPosix = True
-        | otherwise = 
-                not (B.any isBadCharacter x2) &&
-                not (any f $ splitDirectories x2) &&
-                not (isJust (readDriveShare x1) && B.all isPathSeparator x1) &&
-                not (isJust (readDriveUNC x1) && not (hasTrailingPathSeparator x1))
-    where
-        (x1,x2) = splitDrive path
-        f x = map toUpper (B8.unpack $ dropWhileEnd (== 32) $ dropExtensions x) `elem` badElements
-
-{-
+	| B.null path = False
+	| B.elem 0 path = False
+	| isPosix = True
+	| otherwise = 
+		not (B.any isBadCharacter x2) &&
+		not (any f $ splitDirectories x2) &&
+		not (isJust (readDriveShare x1) && B.all isPathSeparator x1) &&
+		not (isJust (readDriveUNC x1) && not (hasTrailingPathSeparator x1))
+  where
+	(x1,x2) = splitDrive path
+	f x = map toUpper (decodeFilePath $ dropWhileEnd (== 32) $ dropExtensions x) `elem` badElements
 
 -- | Take a FilePath and make it valid; does not change already valid FilePaths.
 --
--- > untested isValid (makeValid x)
--- > untested isValid x ==> makeValid x == x
--- > untested makeValid "" == "_"
--- > untested makeValid "file\0name" == "file_name"
--- > untested Windows: makeValid "c:\\already\\/valid" == "c:\\already\\/valid"
--- > untested Windows: makeValid "c:\\test:of_test" == "c:\\test_of_test"
--- > untested Windows: makeValid "test*" == "test_"
--- > untested Windows: makeValid "c:\\test\\nul" == "c:\\test\\nul_"
--- > untested Windows: makeValid "c:\\test\\prn.txt" == "c:\\test\\prn_.txt"
--- > untested Windows: makeValid "c:\\test/prn.txt" == "c:\\test/prn_.txt"
--- > untested Windows: makeValid "c:\\nul\\file" == "c:\\nul_\\file"
--- > untested Windows: makeValid "\\\\\\foo" == "\\\\drive"
--- > untested Windows: makeValid "\\\\?\\D:file" == "\\\\?\\D:\\file"
--- > untested Windows: makeValid "nul .txt" == "nul _.txt"
-makeValid :: FilePath -> FilePath
+-- > isValid (makeValid x)
+-- > isValid x ==> makeValid x == x
+-- > makeValid "" == "_"
+-- > makeValid "file\0name" == "file_name"
+-- > Windows: makeValid "c:\\already\\/valid" == "c:\\already\\/valid"
+-- > Windows: makeValid "c:\\test:of_test" == "c:\\test_of_test"
+-- > Windows: makeValid "test*" == "test_"
+-- > Windows: makeValid "c:\\test\\nul" == "c:\\test\\nul_"
+-- > Windows: makeValid "c:\\test\\prn.txt" == "c:\\test\\prn_.txt"
+-- > Windows: makeValid "c:\\test/prn.txt" == "c:\\test/prn_.txt"
+-- > Windows: makeValid "c:\\nul\\file" == "c:\\nul_\\file"
+-- > Windows: makeValid "\\\\\\foo" == "\\\\drive"
+-- > Windows: makeValid "\\\\?\\D:file" == "\\\\?\\D:\\file"
+-- > Windows: makeValid "nul .txt" == "nul _.txt"
+makeValid :: RawFilePath -> RawFilePath
 makeValid "" = "_"
 makeValid path
-        | isPosix = map (\x -> if x == '\0' then '_' else x) path
-        | isJust (readDriveShare drv) && all isPathSeparator drv = take 2 drv ++ "drive"
+        | isPosix = B.map (\x -> if x == 0 then underscore else x) path
+        | isJust (readDriveShare drv) && B.all isPathSeparator drv = B.take 2 drv <> "drive"
         | isJust (readDriveUNC drv) && not (hasTrailingPathSeparator drv) =
-            makeValid (drv ++ [pathSeparator] ++ pth)
+            makeValid (drv <> B.singleton pathSeparator <> pth)
         | otherwise = joinDrive drv $ validElements $ validChars pth
     where
         (drv,pth) = splitDrive path
 
-        validChars = map f
-        f x = if isBadCharacter x then '_' else x
+	underscore :: Word8
+	underscore = fromIntegral (ord '_')
+
+        validChars = B.map f
+        f x = if isBadCharacter x then underscore else x
 
         validElements x = joinPath $ map g $ splitPath x
-        g x = h a ++ b
-            where (a,b) = break isPathSeparator x
-        h x = if map toUpper (dropWhileEnd (== ' ') a) `elem` badElements then a ++ "_" <.> b else x
+        g x = h a <> b
+            where (a,b) = B.break isPathSeparator x
+        h x = if map toUpper (decodeFilePath $ dropWhileEnd (== 32) a) `elem` badElements then a <> "_" <.> b else x
             where (a,b) = splitExtensions x
 
--}
-
 -- | Is a path relative, or is it fixed to the root?
 --
 -- > Windows: isRelative "path\\test" == True
@@ -1038,7 +1142,8 @@ makeValid path
 -- * "You cannot use the "\\?\" prefix with a relative path."
 isRelative :: RawFilePath -> Bool
 isRelative x = B.null drive || isRelativeDrive drive
-    where drive = takeDrive x
+  where
+	drive = takeDrive x
 
 
 {- c:foo -}
@@ -1047,7 +1152,8 @@ isRelative x = B.null drive || isRelativeDrive drive
 -- current directory on the drive with the specified letter."
 isRelativeDrive :: RawFilePath -> Bool
 isRelativeDrive x =
-    maybe False (not . hasTrailingPathSeparator . fst) (readDriveLetter x)
+	maybe False (not . hasTrailingPathSeparator . fst)
+		(readDriveLetter x)
 
 -- | @not . 'isRelative'@
 --
